/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/ios/input.c: iOS input device interface.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/input.h"
#include "src/sysdep.h"
#include "src/sysdep/ios/input.h"
#include "src/sysdep/ios/osk.h"
#include "src/sysdep/ios/util.h"
#include "src/time.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Event callback passed to sys_input_init(). */
static InputEventCallback event_callback;

/* Flag indicating whether the text input interface is active. */
static uint8_t text_input_on;

/*************************************************************************/
/******************** Interface: Basic functionality *********************/
/*************************************************************************/

int sys_input_init(void (*event_callback_)(const struct InputEvent *))
{
    PRECOND(event_callback_ != NULL, return 0);

    event_callback = event_callback_;
    text_input_on = 0;
    return 1;
}

/*-----------------------------------------------------------------------*/

void sys_input_cleanup(void)
{
    if (ios_osk_is_running()) {
        ios_osk_close();
    }
    event_callback = NULL;
}

/*-----------------------------------------------------------------------*/

void sys_input_update(void)
{
    if (text_input_on && !ios_osk_is_running()) {
        text_input_on = 0;
        InputEvent event = {.type = INPUT_EVENT_TEXT, .timestamp = time_now()};
        const int *text = ios_osk_get_text();
        if (!text) {
            DLOG("Failed to get OSK text");
            event.detail = INPUT_TEXT_CANCELLED;
            (*event_callback)(&event);
        } else {
            event.detail = INPUT_TEXT_CLEAR;
            (*event_callback)(&event);
            event.detail = INPUT_TEXT_INPUT;
            for (; *text; text++) {
                event.text.ch = *text;
                (*event_callback)(&event);
            }
            event.detail = INPUT_TEXT_DONE;
            (*event_callback)(&event);
        }
    }
}

/*-----------------------------------------------------------------------*/

void sys_input_info(SysInputInfo *info_ret)
{
    info_ret->has_joystick    = 0;
    info_ret->has_keyboard    = 0;
    info_ret->has_mouse       = 0;
    info_ret->has_text        = 1;
    info_ret->text_uses_custom_interface = 1;
    info_ret->text_has_prompt = 1;
    info_ret->has_touch       = 1;
}

/*-----------------------------------------------------------------------*/

void sys_input_grab(UNUSED int grab)
{
    /* Meaningless on iOS. */
}

/*-----------------------------------------------------------------------*/

int sys_input_is_quit_requested(void)
{
    return ios_application_is_terminating;
}

/*-----------------------------------------------------------------------*/

int sys_input_is_suspend_requested(void)
{
    return ios_application_is_suspending;
}

/*-----------------------------------------------------------------------*/

void sys_input_acknowledge_suspend_request(void)
{
    sys_semaphore_signal(ios_suspend_semaphore);
    sys_semaphore_wait(ios_resume_semaphore);
}

/*************************************************************************/
/********************* Interface: Joystick handling **********************/
/*************************************************************************/

int sys_input_joystick_button_mapping(UNUSED int index, UNUSED int name)
{
    /* No joystick support. */
    return -1;
}

/*-----------------------------------------------------------------------*/

void sys_input_joystick_rumble(UNUSED int index, UNUSED float left,
                               UNUSED float right, UNUSED float time)
{
    /* No joystick support. */
}

/*************************************************************************/
/*********************** Interface: Mouse handling ***********************/
/*************************************************************************/

void sys_input_mouse_set_position(UNUSED float x, UNUSED float y)
{
    /* No mouse support. */
}

/*************************************************************************/
/******************** Interface: Text entry handling *********************/
/*************************************************************************/

void sys_input_text_set_state(int on, const char *text, const char *prompt)
{
    if (on) {
        ios_osk_open(text ? text : "", prompt);
        text_input_on = 1;
    } else {
        ios_osk_close();
        text_input_on = 0;
    }
}

/*************************************************************************/
/*********************** Library-internal routines ***********************/
/*************************************************************************/

void ios_forward_input_event(const InputEvent *event)
{
    if (event_callback) {
        (*event_callback)(event);
    }
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
