/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/graphics/base.c: Tests for basic graphics functions.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/memory.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/texture.h"

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_graphics_base)

TEST_INIT(init)
{
    return graphics_init();
}

TEST_CLEANUP(cleanup)
{
    graphics_cleanup();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_init_memory_failure)
{
    graphics_cleanup();

    CHECK_MEMORY_FAILURES(graphics_init());

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_init_invalid)
{
    CHECK_FALSE(graphics_init());  // Double init should fail.

    graphics_cleanup();
    graphics_cleanup();  // Double cleanup should be a no-op.

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_sizes)
{
    const GraphicsDisplaySizeList *size_list;
    CHECK_TRUE(size_list = graphics_list_display_sizes());
    CHECK_TRUE(size_list->num_sizes > 0);

    DLOG("Available display sizes:");
    for (int i = 0; i < size_list->num_sizes; i++) {
        DLOG("   %dx%d", size_list->sizes[i].width, size_list->sizes[i].height);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_fullscreen)
{
    const int width = graphics_device_width();
    const int height = graphics_device_height();

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(width, height));
    CHECK_FALSE(graphics_display_is_window());
    CHECK_INTEQUAL(graphics_display_width(), width);
    CHECK_INTEQUAL(graphics_display_height(), height);
    /* We should always have input focus after setting fullscreen mode. */
#ifdef USES_SDL
    /* But on SDL we might have to wait a little while. */
    const uint32_t now = SDL_GetTicks();
    while (!graphics_has_focus() && SDL_GetTicks() - now < 1000) {
        sys_thread_yield();
        SDL_PumpEvents();
        SDL_Event sdl_event;
        while (SDL_PollEvent(&sdl_event)) { /*nothing*/ }
    }
#endif
    CHECK_TRUE(graphics_has_focus());

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_window)
{
    const GraphicsDisplaySizeList *size_list;
    CHECK_TRUE(size_list = graphics_list_display_sizes());
    CHECK_TRUE(size_list->num_sizes > 0);
    const int width = size_list->sizes[0].width;
    const int height = size_list->sizes[0].height;

    if (graphics_has_windowed_mode()) {
        CHECK_TRUE(graphics_set_display_attr("window", 1));
    } else {
        CHECK_FALSE(graphics_set_display_attr("window", 1));
    }
    CHECK_TRUE(graphics_set_display_mode(width, height));
    if (graphics_has_windowed_mode()) {
        CHECK_TRUE(graphics_display_is_window());
    } else {
        CHECK_FALSE(graphics_display_is_window());
    }
    CHECK_INTEQUAL(graphics_display_width(), width);
    CHECK_INTEQUAL(graphics_display_height(), height);

    return 1;
}

/*-----------------------------------------------------------------------*/

/*
 * This test is broken in SDL2:
 * - https://bugzilla.libsdl.org/show_bug.cgi?id=2255
 */
#ifndef USES_SDL2
TEST(test_create_wrong_size)
{
    /* Assume that trying to use an INT_MAX by INT_MAX fullscreen mode
     * will fail.  (I'd like to see a system where it succeeds...) */
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_FALSE(graphics_set_display_mode(INT_MAX, INT_MAX));

    /* Check that we can subsequently open a normal display. */
    const GraphicsDisplaySizeList *size_list;
    CHECK_TRUE(size_list = graphics_list_display_sizes());
    CHECK_TRUE(size_list->num_sizes > 0);
    const int width = size_list->sizes[0].width;
    const int height = size_list->sizes[0].height;
    CHECK_TRUE(graphics_set_display_attr("window",
                                         graphics_has_windowed_mode()));
    CHECK_TRUE(graphics_set_display_mode(width, height));

    return 1;
}
#endif  // USES_SDL2

/*-----------------------------------------------------------------------*/

TEST(test_create_invalid)
{
    const GraphicsDisplaySizeList *size_list;
    CHECK_TRUE(size_list = graphics_list_display_sizes());
    CHECK_TRUE(size_list->num_sizes > 0);
    const int width = size_list->sizes[0].width;
    const int height = size_list->sizes[0].height;

    CHECK_TRUE(graphics_set_display_attr("window",
                                         graphics_has_windowed_mode()));
    CHECK_FALSE(graphics_set_display_mode(0, 0));
    CHECK_FALSE(graphics_set_display_mode(width, 0));
    CHECK_FALSE(graphics_set_display_mode(0, height));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_attr)
{
    const GraphicsDisplaySizeList *size_list;
    CHECK_TRUE(size_list = graphics_list_display_sizes());
    CHECK_TRUE(size_list->num_sizes > 0);
    const int width = size_list->sizes[0].width;
    const int height = size_list->sizes[0].height;

    CHECK_TRUE(graphics_set_display_attr("depth_bits", 16));
    CHECK_TRUE(graphics_set_display_attr("multisample", 1));
    CHECK_TRUE(graphics_set_display_attr("preserve_state", 0));
    CHECK_TRUE(graphics_set_display_attr("stencil_bits", 0));
    /* Both vsync=true and vsync=false could potentially fail depending on
     * system capabilities, so just check that at least one succeeds. */
    CHECK_TRUE(graphics_set_display_attr("vsync", 0)
               || graphics_set_display_attr("vsync", 1));
    CHECK_TRUE(graphics_set_display_attr("window",
                                         graphics_has_windowed_mode()));
    CHECK_TRUE(graphics_set_display_mode(width, height));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_attr_invalid)
{
    const GraphicsDisplaySizeList *size_list;
    CHECK_TRUE(size_list = graphics_list_display_sizes());
    CHECK_TRUE(size_list->num_sizes > 0);
    const int width = size_list->sizes[0].width;
    const int height = size_list->sizes[0].height;

    CHECK_FALSE(graphics_set_display_attr("NO_SUCH_ATTRIBUTE", 0));
    CHECK_FALSE(graphics_set_display_attr("depth_bits", -1));
    CHECK_FALSE(graphics_set_display_attr("multisample", -1));
    CHECK_FALSE(graphics_set_display_attr("stencil_bits", -1));

    CHECK_TRUE(graphics_set_display_attr("window",
                                         graphics_has_windowed_mode()));
    /* The graphics_set_display_mode() call should still succeed (using
     * default attribute values). */
    CHECK_TRUE(graphics_set_display_mode(width, height));

    /* Any of these could succeed if the system can't immediately detect
     * that the values are out of range, but if they do succeed, attempting
     * to set a display mode using those values should fail. */
    if (graphics_set_display_attr("depth_bits", INT_MAX)) {
        CHECK_FALSE(graphics_set_display_mode(width, height));
        CHECK_TRUE(graphics_set_display_attr("depth_bits", 16));
        CHECK_TRUE(graphics_set_display_mode(width, height));
    }
    if (graphics_set_display_attr("multisample", INT_MAX)) {
        CHECK_FALSE(graphics_set_display_mode(width, height));
        CHECK_TRUE(graphics_set_display_attr("multisample", 1));
        CHECK_TRUE(graphics_set_display_mode(width, height));
    }
    if (graphics_set_display_attr("stencil_bits", INT_MAX)) {
        CHECK_FALSE(graphics_set_display_mode(width, height));
        CHECK_TRUE(graphics_set_display_attr("stencil_bits", 0));
        CHECK_TRUE(graphics_set_display_mode(width, height));
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_renderer_info)
{
    const GraphicsDisplaySizeList *size_list;
    CHECK_TRUE(size_list = graphics_list_display_sizes());
    CHECK_TRUE(size_list->num_sizes > 0);
    const int width = size_list->sizes[0].width;
    const int height = size_list->sizes[0].height;

    if (graphics_has_windowed_mode()) {
        CHECK_TRUE(graphics_set_display_attr("window", 1));
    } else {
        CHECK_FALSE(graphics_set_display_attr("window", 1));
    }
    CHECK_TRUE(graphics_set_display_mode(width, height));

    /* We don't know what text we'll get back, but it should never be NULL. */
    CHECK_TRUE(graphics_renderer_info() != NULL);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_window_title)
{
    const GraphicsDisplaySizeList *size_list;
    CHECK_TRUE(size_list = graphics_list_display_sizes());
    CHECK_TRUE(size_list->num_sizes > 0);
    const int width = size_list->sizes[0].width;
    const int height = size_list->sizes[0].height;
    CHECK_TRUE(graphics_set_display_attr("window", graphics_has_windowed_mode()));
    CHECK_TRUE(graphics_set_display_mode(width, height));

    graphics_set_window_title("SIL Test");

    /* We can only check the title on SDL, but SDL is also the only platform
     * that supports window titles to begin with, so on other platforms we
     * just make sure the call doesn't crash. */
#ifdef USES_SDL
# ifdef USES_SDL2
    const char *title = SDL_GetWindowTitle(sdl_window());
    CHECK_STREQUAL(title, "SIL Test");
# else
    const char *title = "", *icon = "";
    SDL_WM_GetCaption((char **)&title, (char **)&icon);
    CHECK_STREQUAL(title, "SIL Test");
    CHECK_STREQUAL(icon, "SIL Test");
# endif
#endif

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_window_title_early)
{
    /* graphics_set_window_title() should work even if it's called before
     * the window is opened. */
    graphics_set_window_title("SIL Test");

    const GraphicsDisplaySizeList *size_list;
    CHECK_TRUE(size_list = graphics_list_display_sizes());
    CHECK_TRUE(size_list->num_sizes > 0);
    const int width = size_list->sizes[0].width;
    const int height = size_list->sizes[0].height;
    CHECK_TRUE(graphics_set_display_attr("window", graphics_has_windowed_mode()));
    CHECK_TRUE(graphics_set_display_mode(width, height));

#if defined(USES_SDL) && !defined(USES_SDL2)
    const char *title = "", *icon = "";
    SDL_WM_GetCaption((char **)&title, (char **)&icon);
    CHECK_STREQUAL(title, "SIL Test");
    CHECK_STREQUAL(icon, "SIL Test");
#endif

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_window_icon)
{
    const GraphicsDisplaySizeList *size_list;
    CHECK_TRUE(size_list = graphics_list_display_sizes());
    CHECK_TRUE(size_list->num_sizes > 0);
    const int width = size_list->sizes[0].width;
    const int height = size_list->sizes[0].height;
    CHECK_TRUE(graphics_set_display_attr("window", graphics_has_windowed_mode()));
    CHECK_TRUE(graphics_set_display_mode(width, height));

    /* We can't retrieve the icon once set, so just make sure it doesn't
     * break the texture. */
    int texture;
    CHECK_TRUE(texture = texture_create(32, 32, 0, 0));
    uint8_t *pixels;
    CHECK_TRUE(pixels = texture_lock(texture));
    memset(pixels, 0xAA, 32*32*4);
    texture_unlock(texture);
    graphics_set_window_icon(texture);
    const uint8_t *pixels_ro;
    CHECK_TRUE(pixels_ro = texture_lock_readonly(texture));
    for (int i = 0; i < 32*32*4; i++) {
        CHECK_INTEQUAL(pixels_ro[i], 0xAA);
    }
    texture_destroy(texture);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_window_title_icon_invalid)
{
    const GraphicsDisplaySizeList *size_list;
    CHECK_TRUE(size_list = graphics_list_display_sizes());
    CHECK_TRUE(size_list->num_sizes > 0);
    const int width = size_list->sizes[0].width;
    const int height = size_list->sizes[0].height;
    CHECK_TRUE(graphics_set_display_attr("window", graphics_has_windowed_mode()));
    CHECK_TRUE(graphics_set_display_mode(width, height));

    /* These don't return values, so just make sure they don't crash. */
    graphics_set_window_title(NULL);
    graphics_set_window_icon(0);
    int texture;
    CHECK_TRUE(texture = texture_create(32, 32, 0, 0));
    texture_destroy(texture);
    graphics_set_window_icon(texture);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_show_mouse_pointer)
{
    /* We have no way of checking whether this works, so just call the
     * function both ways and make sure it doesn't crash. */
    graphics_show_mouse_pointer(1);
    graphics_show_mouse_pointer(0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_get_mouse_pointer_state)
{
#ifdef USES_SDL
    graphics_show_mouse_pointer(1);
    CHECK_TRUE(graphics_get_mouse_pointer_state());
    graphics_show_mouse_pointer(0);
    CHECK_FALSE(graphics_get_mouse_pointer_state());
#else
    /* We have no way of checking whether this works, so just call the
     * function and make sure it doesn't crash. */
    (void) graphics_get_mouse_pointer_state();
#endif

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_get_frame_period)
{
    const double period = graphics_frame_period();
    CHECK_TRUE(period >= 0);

#ifdef USES_SDL
# ifdef USES_SDL2
    /* SDL 2.0 systems should always give us a valid frame period. */
    CHECK_TRUE(period != 0);
# else
    /* SDL 1.2 systems always return 0 because we have no way to get the
     * frame period. */
    CHECK_FLOATEQUAL(period, 0);
# endif
#else
    /* All other platforms return NTSC timing (59.94Hz).  Note that we
     * do the same computation as graphics_frame_period() to ensure
     * equality -- 1.001/60.0 rounds differently from 1001.0/60000.0. */
    volatile int num = 1001, den = 60000; // Prevent constant folding.
    CHECK_FLOATEQUAL(period, (double)num / (double)den);
#endif

    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
