/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/opengl/state.c: Tests for OpenGL render state management.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/internal.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/test/sysdep/opengl/internal.h"
#include "src/texture.h"

/*************************************************************************/
/**************************** Helper routines ****************************/
/*************************************************************************/

/**
 * generate_shader_source:  Shader source generator for testing error paths.
 * The shader defines the "transform" uniform incorrectly, so attempts to
 * set it in opengl_apply_shader() will raise GL errors.
 */
static int generate_shader_source(
    UNUSED GLuint vertex_shader, UNUSED GLuint fragment_shader,
    UNUSED GLenum primitive_type, UNUSED int position_count,
    UNUSED int texcoord_count, UNUSED GLenum texcolor_type,
    UNUSED int tex_offset, UNUSED int color_count, UNUSED int color_uniform,
    UNUSED int fog, UNUSED int alpha_test, UNUSED GLenum alpha_comparison)
{
    #ifdef SIL_OPENGL_SHADER_VERSION
    # define SIL_OPENGL_SHADER_VERSION_STRING _STR1(SIL_OPENGL_SHADER_VERSION)
    # define _STR1(x) _STR2(x)
    # define _STR2(x) #x
    # define SHADER_VERSION_LINE_ "#version "SIL_OPENGL_SHADER_VERSION_STRING"\n",
    #else
    # define SHADER_VERSION_LINE_ /*nothing*/
    #endif

    #if defined(SIL_OPENGL_SHADER_VERSION) && SIL_OPENGL_SHADER_VERSION >= 130
    # define VERTEX_IN    "in"
    # define GL_FRAGCOLOR "frag_color"
    #else
    # define VERTEX_IN    "attribute"
    # define GL_FRAGCOLOR "gl_FragColor"
    #endif

    #ifdef SIL_OPENGL_ES
    # define HIGHP   "highp"
    # define LOWP    "lowp"
    #else
    # define HIGHP   ""
    # define LOWP    ""
    #endif

    static const char *vs_source[] = {
        SHADER_VERSION_LINE_
        "uniform int transform;\n",
        VERTEX_IN" "HIGHP" vec4 position;\n",
        "void main() {\n",
        "    gl_Position = position * float(transform);\n",
        "}\n",
    };
    static const char *fs_source[] = {
        SHADER_VERSION_LINE_
#if defined(SIL_OPENGL_SHADER_VERSION) && SIL_OPENGL_SHADER_VERSION >= 130
        "out "LOWP" vec4 frag_color;\n",
#endif
        "void main() {\n",
        "    "GL_FRAGCOLOR" = vec4(1.0, 1.0, 1.0, 1.0);\n",
        "}\n",
    };

    glShaderSource(vertex_shader, lenof(vs_source), vs_source, NULL);
    glShaderSource(fragment_shader, lenof(fs_source), fs_source, NULL);

    return 1;

    #undef VERTEX_IN
    #undef GL_FRAGCOLOR
    #undef HIGHP
}

/*-----------------------------------------------------------------------*/

/**
 * generate_shader_key:  Shader key generator for testing error paths.
 */
static uint32_t generate_shader_key(
    UNUSED GLenum primitive_type, UNUSED int position_count,
    UNUSED int texcoord_count, UNUSED GLenum texcolor_type,
    UNUSED int tex_offset, UNUSED int color_count, UNUSED int color_uniform,
    UNUSED int fog, UNUSED int alpha_test, UNUSED GLenum alpha_comparison)
{
    return 0;
}

/*************************************************************************/
/****************************** Test runner ******************************/
/*************************************************************************/

static int do_test_opengl_state(void);
int test_opengl_state(void)
{
    if (!run_tests_in_window(do_test_opengl_state)) {
        return 0;
    }

    if (opengl_has_features_uninitted(OPENGL_FEATURE_SHADERS
                                      | OPENGL_FEATURE_FFP_PRIMITIVES)) {
        TEST_opengl_force_feature_mask = OPENGL_FEATURE_SHADERS;
        TEST_opengl_force_feature_flags = 0;
        const int result = run_tests_in_window(do_test_opengl_state);
        TEST_opengl_force_feature_mask = 0;
        TEST_opengl_force_feature_flags = 0;
        if (!result) {
            FAIL("Preceding failure(s) occurred with shaders disabled");
        }
    }

    return 1;
}

DEFINE_GENERIC_TEST_RUNNER(do_test_opengl_state)

TEST_INIT(init)
{
    graphics_start_frame();
    graphics_set_viewport(0, 0, 64, 64);
    graphics_clear(0, 0, 0);
    return 1;
}

TEST_CLEANUP(cleanup)
{
    graphics_enable_alpha_test(0);
    graphics_set_alpha_test_comparison(GRAPHICS_COMPARISON_GREATER_EQUAL);
    graphics_set_alpha_test_reference(0);
    graphics_set_blend(GRAPHICS_BLEND_ADD,
                       GRAPHICS_BLEND_SRC_ALPHA, GRAPHICS_BLEND_INV_SRC_ALPHA);
    graphics_set_blend_color(&(Vector4f){0,0,0,0});
    graphics_set_clip_region(0, 0, 0, 0);
    graphics_enable_depth_test(0);
    graphics_set_depth_test_comparison(GRAPHICS_COMPARISON_LESS);
    graphics_enable_depth_write(1);
    graphics_set_depth_range(0, 1);
    graphics_set_face_cull(GRAPHICS_FACE_CULL_NONE);
    graphics_enable_fog(0);
    graphics_set_fog_start(0);
    graphics_set_fog_end(0);
    graphics_set_fog_color(&(Vector4f){1,1,1,1});
    graphics_enable_stencil_test(0);
    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_TRUE, 0, ~0U);
    graphics_set_stencil_operations(GRAPHICS_STENCIL_KEEP,
                                    GRAPHICS_STENCIL_KEEP,
                                    GRAPHICS_STENCIL_KEEP);
    texture_apply(0);

    graphics_finish_frame();
    graphics_flush_resources();
    return 1;
}

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

TEST(test_restore_alpha_test)
{
    if (!opengl_has_features(OPENGL_FEATURE_FFP_PRIMITIVES)) {
        SKIP("Fixed-function pipeline not available.");
    }

    graphics_enable_alpha_test(1);
    graphics_set_alpha_test_reference(0.5);
    draw_square(0, 1, 1, 1, 0.2);
    CHECK_SQUARE(0, 0, 0);

    opengl_save_state();
    glDisable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GEQUAL, 0.1);
    opengl_restore_state();
    draw_square(0, 1, 1, 1, 0.2);
    CHECK_SQUARE(0, 0, 0);

    graphics_set_alpha_test_comparison(GRAPHICS_COMPARISON_LESS);
    graphics_set_alpha_test_reference(0.1);
    draw_square(0, 1, 1, 1, 0.2);
    CHECK_SQUARE(0, 0, 0);

    opengl_save_state();
    glDisable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GEQUAL, 0.1);
    opengl_restore_state();
    draw_square(0, 1, 1, 1, 0.2);
    CHECK_SQUARE(0, 0, 0);

    graphics_enable_alpha_test(0);
    draw_square(0, 1, 1, 1, 0.2);
    CHECK_SQUARE(0.2, 0.2, 0.2);

    opengl_save_state();
    glEnable(GL_ALPHA_TEST);
    opengl_restore_state();
    draw_square(0, 1, 1, 1, 0.25);
    CHECK_SQUARE(0.4, 0.4, 0.4);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_blend_enable)
{
    graphics_set_blend(GRAPHICS_BLEND_ADD,
                       GRAPHICS_BLEND_ONE, GRAPHICS_BLEND_ZERO);
    draw_square(0, 1, 0, 0, 0.2);
    CHECK_SQUARE(1, 0, 0);

    opengl_save_state();
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    opengl_restore_state();
    draw_square(0, 0, 1, 0, 0.2);
    CHECK_SQUARE(0, 1, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_blend_func)
{
    graphics_set_blend(GRAPHICS_BLEND_ADD,
                       GRAPHICS_BLEND_ONE, GRAPHICS_BLEND_ZERO);
    draw_square(0, 0.2, 0.4, 0.6, 1);
    CHECK_SQUARE(0.2, 0.4, 0.6);

    graphics_set_blend(GRAPHICS_BLEND_ADD,
                       GRAPHICS_BLEND_ONE, GRAPHICS_BLEND_ONE);
    draw_square(0, 0.2, 0.2, 0.2, 1);
    CHECK_SQUARE(0.4, 0.6, 0.8);

    opengl_save_state();
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    opengl_restore_state();
    draw_square(0, 0.2, 0.2, 0.2, 1);
    CHECK_SQUARE(0.6, 0.8, 1);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_blend_color)
{
    graphics_set_blend(GRAPHICS_BLEND_ADD,
                       GRAPHICS_BLEND_ONE, GRAPHICS_BLEND_ZERO);
    draw_square(0, 0.6, 0.6, 0.6, 1);
    CHECK_SQUARE(0.6, 0.6, 0.6);

    graphics_set_blend(GRAPHICS_BLEND_ADD,
                       GRAPHICS_BLEND_ONE, GRAPHICS_BLEND_CONSTANT);
    graphics_set_blend_color(&(Vector4f){0, 1.0/3.0, 2.0/3.0, 1});
    draw_square(0, 0.2, 0.4, 0.6, 1);
    CHECK_SQUARE(0.2, 0.6, 1);

    opengl_save_state();
    glBlendColor(1, 1, 1, 1);
    opengl_restore_state();
    draw_square(0, 0.6, 0.4, 0.2, 1);
    CHECK_SQUARE(0.6, 0.6, 0.2+2.0/3.0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_clip_enable)
{
    graphics_set_clip_region(0, 0, 1, 1);
    draw_square(0, 1, 1, 1, 1);
    CHECK_SQUARE(0, 0, 0);

    opengl_save_state();
    glDisable(GL_SCISSOR_TEST);
    opengl_restore_state();
    draw_square(0, 1, 1, 1, 1);
    CHECK_SQUARE(0, 0, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_depth_test)
{
    graphics_enable_depth_test(1);
    draw_square(0, 0, 1, 0, 1);
    CHECK_SQUARE(0, 1, 0);

    opengl_save_state();
    glDisable(GL_DEPTH_TEST);
    opengl_restore_state();
    draw_square(1, 1, 0, 0, 1);
    CHECK_SQUARE(0, 1, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_depth_func)
{
    graphics_enable_depth_test(1);
    graphics_set_depth_test_comparison(GRAPHICS_COMPARISON_LESS_EQUAL);
    draw_square(0, 1, 0, 0, 1);
    CHECK_SQUARE(1, 0, 0);

    opengl_save_state();
    glDepthFunc(GL_LESS);
    opengl_restore_state();
    draw_square(0, 0, 0, 1, 1);
    CHECK_SQUARE(0, 0, 1);

    graphics_set_depth_test_comparison(GRAPHICS_COMPARISON_LESS);
    draw_square(-1, 0, 1, 0, 1);
    CHECK_SQUARE(0, 1, 0);

    opengl_save_state();
    glDepthFunc(GL_LEQUAL);
    opengl_restore_state();
    draw_square(-1, 1, 0, 0, 1);
    CHECK_SQUARE(0, 1, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_depth_mask)
{
    graphics_enable_depth_test(1);
    graphics_enable_depth_write(0);
    draw_square(0, 1, 0, 0, 1);
    draw_square(0.5, 0, 0, 1, 1);
    CHECK_SQUARE(0, 0, 1);

    opengl_save_state();
    glDepthMask(GL_TRUE);
    opengl_restore_state();
    draw_square(0, 1, 0, 0, 1);
    draw_square(0.5, 0, 1, 0, 1);
    CHECK_SQUARE(0, 1, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_depth_range)
{
    graphics_enable_depth_test(1);
    graphics_set_depth_range(0.75, 1);
    draw_square(0, 0, 1, 0, 1);  // Has a depth value of 0.875.
    CHECK_SQUARE(0, 1, 0);

    opengl_save_state();
    glDepthRangef(0, 1);
    opengl_restore_state();
    draw_square(0.2, 1, 0, 0, 1);  // Would have depth 0.6 under default range.
    CHECK_SQUARE(0, 1, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_face_cull)
{
    graphics_set_face_cull(GRAPHICS_FACE_CULL_CCW);
    /* Assume draw_square() draws vertices in clockwise order. */
    draw_square(0, 1, 0, 0, 1);
    CHECK_SQUARE(1, 0, 0);

    opengl_save_state();
    glFrontFace(GL_CCW);
    opengl_restore_state();
    draw_square(1, 0, 1, 0, 1);
    CHECK_SQUARE(0, 1, 0);

    graphics_set_face_cull(GRAPHICS_FACE_CULL_CW);
    draw_square(0, 0, 0, 1, 1);
    CHECK_SQUARE(0, 1, 0);

    opengl_save_state();
    glFrontFace(GL_CW);
    opengl_restore_state();
    draw_square(1, 1, 0, 0, 1);
    CHECK_SQUARE(0, 1, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_fog)
{
    if (!opengl_has_features(OPENGL_FEATURE_FFP_PRIMITIVES)) {
        SKIP("Fixed-function pipeline not available.");
    }

    uint8_t *pixels;

    graphics_set_viewport(0, 0, 65, 65);  // Intentionally 65.

    graphics_enable_fog(1);
    graphics_set_fog_start(0.25);
    graphics_set_fog_end(0.875);
    graphics_set_fog_color(&(Vector4f){0, 1.0/3.0, 2.0/3.0, 1});
    CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_POINTS));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){0, 0, 0.5}, NULL, NULL));
    CHECK_TRUE(graphics_end_and_draw_primitive());
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        const int x = (i/4) % TESTW;
        const int y = (i/4) / TESTW;
        const int r = (x==32 && y==32 ? 153 : 0);
        const int g = (x==32 && y==32 ? 187 : 0);
        const int b = (x==32 && y==32 ? 221 : 0);
        CHECK_PIXEL(&pixels[i], r,g,b,255, x, y);
    }
    mem_free(pixels);

    opengl_save_state();
    glDisable(GL_FOG);
    glFogf(GL_FOG_START, 0);
    glFogf(GL_FOG_END, 1);
    glFogfv(GL_FOG_COLOR, (float[4]){1, 1, 1, 1});
    opengl_restore_state();
    CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_POINTS));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){0, 0, 0.625}, NULL, NULL));
    CHECK_TRUE(graphics_end_and_draw_primitive());
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        const int x = (i/4) % TESTW;
        const int y = (i/4) / TESTW;
        const int r = (x==32 && y==32 ? 102 : 0);
        const int g = (x==32 && y==32 ? 153 : 0);
        const int b = (x==32 && y==32 ? 204 : 0);
        CHECK_PIXEL_NEAR(&pixels[i], r,g,b,255, 1, x, y);
    }
    mem_free(pixels);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_stencil)
{
    graphics_enable_stencil_test(1);
    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_TRUE, 200, 255);
    graphics_set_stencil_operations(GRAPHICS_STENCIL_REPLACE,
                                    GRAPHICS_STENCIL_REPLACE,
                                    GRAPHICS_STENCIL_REPLACE);
    draw_square(0, 1,1,1,1);

    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_NOT_EQUAL, 200, 255);
    graphics_set_stencil_operations(GRAPHICS_STENCIL_DECR,
                                    GRAPHICS_STENCIL_KEEP,
                                    GRAPHICS_STENCIL_KEEP);
    opengl_save_state();
    glDisable(GL_STENCIL_TEST);
    glStencilFunc(GL_GREATER, 222, 255);
    glStencilOp(GL_INCR, GL_REPLACE, GL_REPLACE);
    opengl_restore_state();
    draw_square(0, 0.2, 0.4, 0.6, 1);  // Should decrement all stencil values.
    CHECK_SQUARE(1, 1, 1);

    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_LESS, 200, 255);
    graphics_set_stencil_operations(GRAPHICS_STENCIL_DECR,
                                    GRAPHICS_STENCIL_REPLACE,
                                    GRAPHICS_STENCIL_REPLACE);
    opengl_save_state();
    glDisable(GL_STENCIL_TEST);
    glStencilFunc(GL_GREATER, 222, 255);
    glStencilOp(GL_INCR, GL_KEEP, GL_KEEP);
    opengl_restore_state();
    draw_square(0, 0.2, 0.4, 0.6, 1);  // Should replace all values with 200.
    CHECK_SQUARE(0.2, 0.4, 0.6);

    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_EQUAL, 200, 255);
    graphics_set_stencil_operations(GRAPHICS_STENCIL_KEEP,
                                    GRAPHICS_STENCIL_KEEP,
                                    GRAPHICS_STENCIL_KEEP);
    draw_square(0, 0.4, 0.6, 0.2, 1);
    CHECK_SQUARE(0.4, 0.6, 0.2);

    graphics_enable_stencil_test(0);
    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_NOT_EQUAL, 200, 255);
    graphics_set_stencil_operations(GRAPHICS_STENCIL_KEEP,
                                    GRAPHICS_STENCIL_KEEP,
                                    GRAPHICS_STENCIL_KEEP);
    opengl_save_state();
    glEnable(GL_STENCIL_TEST);
    opengl_restore_state();
    draw_square(0, 0.6, 0.2, 0.4, 1);
    CHECK_SQUARE(0.6, 0.2, 0.4);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_texture)
{
    static const uint8_t texture_data[] = {0x33, 0x66, 0x99, 0xFF};
    int texture;
    CHECK_TRUE(texture = texture_create(1, 1, 0, 0));
    uint8_t *texture_pixels;
    CHECK_TRUE(texture_pixels = texture_lock_writeonly(texture));
    memcpy(texture_pixels, texture_data, sizeof(texture_data));
    texture_unlock(texture);

    struct TexturedVertex {float x,y,z,u,v;};
    static const struct TexturedVertex tex_quad_vertices[] = {
        {-0.5,-0.5,0,0,0}, {-0.5,0.5,0,0,1}, {0.5,0.5,0,1,1}, {0.5,-0.5,0,1,0}
    };
    static const uint32_t textured_vertex_format[] = {
        GRAPHICS_VERTEX_FORMAT(POSITION_3F, offsetof(struct TexturedVertex,x)),
        GRAPHICS_VERTEX_FORMAT(TEXCOORD_2F, offsetof(struct TexturedVertex,u)),
        0
    };

    graphics_set_blend(GRAPHICS_BLEND_ADD,
                       GRAPHICS_BLEND_ONE, GRAPHICS_BLEND_ONE);
    texture_apply(texture);
    graphics_draw_vertices(
        GRAPHICS_PRIMITIVE_QUADS, tex_quad_vertices, textured_vertex_format,
        sizeof(*tex_quad_vertices), lenof(tex_quad_vertices));
    CHECK_SQUARE(0.2, 0.4, 0.6);

    /* opengl_save_state() will destroy existing textures, so we call
     * opengl_state_reapply() directly to verify that it restores the
     * current texture ID. */
    glBindTexture(GL_TEXTURE_2D, 0);
    if (!opengl_has_features(OPENGL_FEATURE_SHADERS)) {
        glDisable(GL_TEXTURE_2D);
    }
    opengl_state_reapply();
    graphics_draw_vertices(
        GRAPHICS_PRIMITIVE_QUADS, tex_quad_vertices, textured_vertex_format,
        sizeof(*tex_quad_vertices), lenof(tex_quad_vertices));
    CHECK_SQUARE(0.4, 0.8, 1);

    texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_apply_texture_target)
{
#ifndef GL_TEXTURE_3D  // Only GLES 2.x on iOS and desktop GL 1.0/1.1.
    DLOG("Note: GL_TEXTURE_3D not supported, can't run test");
#else
    static const uint8_t texture_data[] = {0x33, 0x66, 0x99, 0xFF};
    int texture;
    CHECK_TRUE(texture = texture_create(1, 1, 0, 0));
    uint8_t *texture_pixels;
    CHECK_TRUE(texture_pixels = texture_lock_writeonly(texture));
    memcpy(texture_pixels, texture_data, sizeof(texture_data));
    texture_unlock(texture);

    struct TexturedVertex {float x,y,z,u,v;};
    static const struct TexturedVertex tex_quad_vertices[] = {
        {-0.5,-0.5,0,0,0}, {-0.5,0.5,0,0,1}, {0.5,0.5,0,1,1}, {0.5,-0.5,0,1,0}
    };
    static const uint32_t textured_vertex_format[] = {
        GRAPHICS_VERTEX_FORMAT(POSITION_3F, offsetof(struct TexturedVertex,x)),
        GRAPHICS_VERTEX_FORMAT(TEXCOORD_2F, offsetof(struct TexturedVertex,u)),
        0
    };

    graphics_set_blend(GRAPHICS_BLEND_ADD,
                       GRAPHICS_BLEND_ONE, GRAPHICS_BLEND_ONE);
    /* We don't support 3D textures, but this exercises the changed-target
     * path in opengl_apply_texture() without requiring the ES-specific
     * TEXTURE_EXTERNAL. */
    opengl_apply_texture(GL_TEXTURE_3D, 1);
    opengl_clear_error();  // Since the glBindTexture() call will have failed.
    texture_apply(texture);
    graphics_draw_vertices(
        GRAPHICS_PRIMITIVE_QUADS, tex_quad_vertices, textured_vertex_format,
        sizeof(*tex_quad_vertices), lenof(tex_quad_vertices));
    CHECK_SQUARE(0.2, 0.4, 0.6);

    texture_destroy(texture);
#endif  // GL_TEXTURE_3D

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_apply_shader_error)
{
    if (!opengl_has_features(OPENGL_FEATURE_SHADERS)) {
        SKIP("Shaders not available.");
    }

    opengl_set_shader_callbacks(generate_shader_source, generate_shader_key,
                                1, 0);

    SysPrimitive primitive;
    primitive.type = GL_TRIANGLES;
    primitive.position_size = 3;
    primitive.texcoord_size = 0;
    primitive.color_size = 0;
    CHECK_FALSE(opengl_apply_shader(&primitive));

    opengl_set_shader_callbacks(NULL, NULL, 0, 0);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_apply_shader_invalid)
{
    if (!opengl_has_features(OPENGL_FEATURE_SHADERS)) {
        SKIP("Shaders not available.");
    }

    SysPrimitive primitive;
    primitive.type = GL_TRIANGLES;
    primitive.position_size = 3;
    primitive.texcoord_size = 0;
    primitive.color_size = 0;
    CHECK_TRUE(opengl_apply_shader(&primitive));

    primitive.position_size = 5;  // Deliberately invalid.
    CHECK_FALSE(opengl_apply_shader(&primitive));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_blend_factor_invalid)
{
    graphics_set_blend(GRAPHICS_BLEND_ADD,
                       GRAPHICS_BLEND_ONE, GRAPHICS_BLEND_ZERO);
    draw_square(0, 0.2, 0.4, 0.6, 1);
    CHECK_SQUARE(0.2, 0.4, 0.6);

    /* Call sys_*() directly to bypass checks in graphics_set_blend(). */
    sys_graphics_set_blend(-1, -1, -1);
    /* The error return of 0 from convert_blend_factor() is GL_ZERO, which
     * is a valid blend factor. */
    CHECK_INTEQUAL(glGetError(), GL_NO_ERROR);
    draw_square(0, 0.2, 0.2, 0.2, 1);
    CHECK_SQUARE(0, 0, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_stencil_invalid)
{
    graphics_enable_stencil_test(1);
    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_TRUE, 200, 255);
    graphics_set_stencil_operations(GRAPHICS_STENCIL_REPLACE,
                                    GRAPHICS_STENCIL_REPLACE,
                                    GRAPHICS_STENCIL_REPLACE);
    draw_square(0, 1,1,1,1);
    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_EQUAL, 200, 255);
    graphics_set_stencil_operations(GRAPHICS_STENCIL_DECR,
                                    GRAPHICS_STENCIL_DECR,
                                    GRAPHICS_STENCIL_DECR);
    draw_square(0, 0.2, 0.4, 0.6, 1);
    CHECK_SQUARE(0.2, 0.4, 0.6);

    /* Call sys_*() directly to bypass checks in graphics_*(). */
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_STENCIL_COMPARISON, -1);
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_STENCIL_REFERENCE, 199);
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_STENCIL_MASK, 255);
    CHECK_INTEQUAL(glGetError(), GL_INVALID_ENUM);
    draw_square(0, 0.4, 0.6, 0.2, 1);
    CHECK_SQUARE(0.2, 0.4, 0.6);
    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_EQUAL, 198, 255);
    draw_square(0, 0.4, 0.6, 0.2, 1);
    CHECK_SQUARE(0.4, 0.6, 0.2);

    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_EQUAL, 197, 255);
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_STENCIL_OP_SFAIL, -1);
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_STENCIL_OP_DFAIL, -1);
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_STENCIL_OP_DPASS, -1);
    /* The error return of 0 is GL_ZERO, so stencil buffer will be zeroed. */
    CHECK_INTEQUAL(glGetError(), GL_NO_ERROR);
    draw_square(0, 0.6, 0.2, 0.4, 1);
    CHECK_SQUARE(0.6, 0.2, 0.4);
    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_EQUAL, 0, 255);
    draw_square(0, 0.2, 0.4, 0.6, 1);
    CHECK_SQUARE(0.2, 0.4, 0.6);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_invalid_param)
{
    draw_square(0, 1, 0, 0, 1);
    CHECK_SQUARE(1, 0, 0);

    const int INVALID_PARAM = -9999;  // Must not match any in sysdep.h.

    /* Call sys_*() directly to bypass checks in graphics_*(). */
    sys_graphics_set_int_param(INVALID_PARAM, 0);
    sys_graphics_set_float_param(INVALID_PARAM, 0);
    sys_graphics_set_vec2_param(INVALID_PARAM, &(Vector2f){0, 0});
    sys_graphics_set_vec4_param(INVALID_PARAM, &(Vector4f){0, 0, 0, 0});
    sys_graphics_set_matrix_param(
        INVALID_PARAM, &(Matrix4f){0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0});

    Matrix4f m = {-1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1};
    sys_graphics_get_matrix_param(INVALID_PARAM, &m);
    CHECK_FLOATEQUAL(m._11, -1);
    CHECK_FLOATEQUAL(m._12, -1);
    CHECK_FLOATEQUAL(m._13, -1);
    CHECK_FLOATEQUAL(m._14, -1);
    CHECK_FLOATEQUAL(m._21, -1);
    CHECK_FLOATEQUAL(m._22, -1);
    CHECK_FLOATEQUAL(m._23, -1);
    CHECK_FLOATEQUAL(m._24, -1);
    CHECK_FLOATEQUAL(m._31, -1);
    CHECK_FLOATEQUAL(m._32, -1);
    CHECK_FLOATEQUAL(m._33, -1);
    CHECK_FLOATEQUAL(m._34, -1);
    CHECK_FLOATEQUAL(m._41, -1);
    CHECK_FLOATEQUAL(m._42, -1);
    CHECK_FLOATEQUAL(m._43, -1);
    CHECK_FLOATEQUAL(m._44, -1);

    draw_square(0, 0, 1, 0, 1);
    CHECK_SQUARE(0, 1, 0);

    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
